import datetime
from dataclasses import dataclass
from typing import Literal, Optional, Sequence

from google.protobuf.timestamp_pb2 import Timestamp

from .proto import chat_pb2

SearchMode = Literal["auto", "on", "off"]


@dataclass
class SearchParameters:
    """Parameters for configuring search behavior in a chat request.

    This class allows customization of search functionality when using models that support
    searching external sources for information. You can specify which sources to search,
    set date ranges for relevant content, control the search mode, and configure how
    results are returned.

    Attributes:
        sources: A sequence of search sources to query, such as web, news, X (Twitter), or RSS feeds.
            Use helper functions like `web_source()`, `news_source()`, `x_source()`, or `rss_source()` to
            create these sources. Multiple sources can be specified. If no sources are specified, the model will
            use the default search sources which are web and X.
        mode: Controls when search is performed. Possible values are:
            - "auto": The model decides when to perform a search based on the prompt(default).
            - "on": Search is always performed when sampling from the model.
            - "off": Search is disabled.
        from_date: Optional start date for search results. Only content after this date
            will be considered. Defaults to None (no start date restriction).
        to_date: Optional end date for search results. Only content before this date
            will be considered. Defaults to None (no end date restriction).
        return_citations: If True, the model will return a list of citations (the domains of the sources used)
            or references to the sources used in generating the response. Defaults to True.
        max_search_results: Optional limit on the number of search results to consider
            when generating a response. Defaults to 15.

    Example:
        ```
        from xai_sdk.chat import SearchParameters, web_source, news_source

        # Configure search to use web and news sources with citations
        search_config = SearchParameters(
            sources=[web_source(country="US"), news_source(country="US")],
            mode="on",
            return_citations=True,
            max_search_results=10
        )

        # Use in a chat request
        chat = client.chat.create(
            model="grok-3-latest",
            messages=[user("What's the latest news on AI developments?")],
            search_parameters=search_config
        )
        ```
    """

    sources: Optional[Sequence[chat_pb2.Source]] = None
    mode: SearchMode = "auto"
    from_date: Optional[datetime.datetime] = None
    to_date: Optional[datetime.datetime] = None
    return_citations: bool = True
    max_search_results: Optional[int] = None

    def _to_proto(self) -> chat_pb2.SearchParameters:
        """Converts the search parameters to a chat_pb2.SearchParameters proto."""
        return chat_pb2.SearchParameters(
            from_date=Timestamp().FromDatetime(self.from_date) if self.from_date else None,
            to_date=Timestamp().FromDatetime(self.to_date) if self.to_date else None,
            max_search_results=self.max_search_results,
            return_citations=self.return_citations,
            mode=self._search_mode_to_proto(self.mode),
            sources=self.sources,
        )

    def _search_mode_to_proto(self, mode: SearchMode) -> chat_pb2.SearchMode:
        match mode:
            case "auto":
                return chat_pb2.SearchMode.AUTO_SEARCH_MODE
            case "on":
                return chat_pb2.SearchMode.ON_SEARCH_MODE
            case "off":
                return chat_pb2.SearchMode.OFF_SEARCH_MODE
            case _:
                raise ValueError(f"Invalid search mode: {mode}. Must be one of: {SearchMode.__args__}")


def web_source(
    country: Optional[str] = None,
    excluded_websites: Optional[Sequence[str]] = None,
    allowed_websites: Optional[Sequence[str]] = None,
    *,
    safe_search: bool = True,
) -> chat_pb2.Source:
    """Creates a web source configuration for search requests.

    This function configures a web source for searching online content, it should be passed to
    the `sources` parameter of the `SearchParameters` class.

    You can specify a country to focus the search on region-specific results and exclude specific websites
    to avoid unwanted sources.

    Args:
        country: Optional ISO country code (e.g., "US", "UK") to limit search results to
            content from a specific region or country. Defaults to None (global search).
        excluded_websites: Optional list of website domains (without the protocol prefix) to exclude
            from search results. e.g. `["example.com"]`.
            Use this to prevent results from specific sites. Defaults to None (no exclusions).
            A maximum of 5 websites can be excluded.
            Note: This parameter cannot be set together with `allowed_websites`.
        allowed_websites: Optional list of website domains (without the protocol prefix) to restrict
            search results to. e.g. `["example.com"]`.
            Use this to limit results to only these specific sites; no other websites will be considered
            in the search process. Defaults to None (no restrictions).
            A maximum of 5 websites can be allowed.
            Note: If no relevant information is found on these websites, the number of results returned
            might be smaller than `max_search_results` configured in `SearchParameters`.
            Note: This parameter cannot be set together with `excluded_websites`.
        safe_search: Optional boolean to enable or disable safe search. Defaults to True (enabled).

    Returns:
        A `chat_pb2.Source` object configured for web search.

    Example:
        ```
        from xai_sdk.chat import SearchParameters, web_source

        search_config = SearchParameters(
            sources=[web_source(country="US", excluded_websites=["example.com"])],
            mode="on"
        )
        ```
    """
    return chat_pb2.Source(
        web=chat_pb2.WebSource(
            country=country,
            excluded_websites=excluded_websites,
            allowed_websites=allowed_websites,
            safe_search=safe_search,
        )
    )


def news_source(
    country: Optional[str] = None,
    excluded_websites: Optional[Sequence[str]] = None,
    *,
    safe_search: bool = True,
) -> chat_pb2.Source:
    """Creates a news source configuration for search requests.

    This function configures a news source for searching recent articles and reports from
    news sources. It should be passed to the `sources` parameter of the
    `SearchParameters` class.

    You can limit results to a specific country for more relevant news content.

    Args:
        country: Optional ISO country code (e.g., "US", "UK") to limit search results to
            news from a specific region or country. Defaults to None (global news).
        excluded_websites: Optional list of website domains (without the protocol prefix) to exclude
            from search results. e.g. `["example.com"]`.
            Use this to prevent results from specific sites. Defaults to None (no exclusions).
            A maximum of 5 websites can be excluded.
        safe_search: Optional boolean to enable or disable safe search. Defaults to True (enabled).

    Returns:
        A `chat_pb2.Source` object configured for news search.

    Example:
        ```
        from xai_sdk.chat import SearchParameters, news_source

        search_config = SearchParameters(
            sources=[news_source(country="US")],
            mode="on"
        )
        ```
    """
    return chat_pb2.Source(
        news=chat_pb2.NewsSource(country=country, excluded_websites=excluded_websites, safe_search=safe_search)
    )


def x_source(
    included_x_handles: Optional[Sequence[str]] = None,
    excluded_x_handles: Optional[Sequence[str]] = None,
    post_favorite_count: Optional[int] = None,
    post_view_count: Optional[int] = None,
) -> chat_pb2.Source:
    """Creates an X (Twitter) source configuration for search requests.

    This function configures a source for searching content on X (formerly Twitter).
    It should be passed to the `sources` parameter of the `SearchParameters` class.

    You can specify particular handles to focus the search on specific users' posts and interactions.

    Args:
        included_x_handles: Optional list of X usernames (without the '@' symbol) to limit search
            results to posts from specific accounts. Defaults to None (general X content).
        excluded_x_handles: Optional list of X usernames (without the '@' symbol) to exclude
            from search results. Defaults to None (no exclusions).
        post_favorite_count: Optional post favorite count threshold. Only posts with at least this many favorites
            will be considered. Defaults to None (no filtering).
        post_view_count: Optional post view count threshold. Only posts with at least this many views
            will be considered. Defaults to None (no filtering).

    Returns:
        A `chat_pb2.Source` object configured for X search.

    Example:
        ```
        from xai_sdk.chat import SearchParameters, x_source

        search_config = SearchParameters(
            sources=[x_source(included_x_handles=["xai"])],
            mode="on"
        )
        ```
    """
    return chat_pb2.Source(
        x=chat_pb2.XSource(
            included_x_handles=included_x_handles,
            excluded_x_handles=excluded_x_handles,
            post_favorite_count=post_favorite_count,
            post_view_count=post_view_count,
        )
    )


def rss_source(links: Sequence[str]) -> chat_pb2.Source:
    """Creates an RSS source configuration for search requests.

    This function configures an RSS source for searching content from RSS feeds.
    It should be passed to the `sources` parameter of the `SearchParameters` class.

    Args:
        links: List of RSS feed URLs to search.

    Returns:
        A `chat_pb2.Source` object configured for RSS search.
    """
    return chat_pb2.Source(rss=chat_pb2.RssSource(links=links))
